def concordance_index(event_times, predicted_times, event_observed):
    def valid_comparison(time_a, time_b, event_a, event_b):
        if time_a == time_b:
            return event_a != event_b
        elif event_a and event_b:
            return True
        elif event_a and time_a < time_b:
            return True
        elif event_b and time_b < time_a:
            return True
        else:
            return False

    def concordance_value(time_a, time_b, pred_a, pred_b):
        if pred_a == pred_b:
            return 0.5
        elif pred_a < pred_b:
            return (time_a < time_b) or (time_a == time_b and event_a and not event_b)
        else:
            return (time_a > time_b) or (time_a == time_b and not event_a and event_b)

    paircount = 0.0
    csum = 0.0

    for a in range(0,len(event_times)):
        time_a = event_times[a]
        pred_a = predicted_times[a]
        event_a = event_observed[a]
        for b in range(a + 1, len(event_times)):
            time_b = event_times[b]
            pred_b = predicted_times[b]
            event_b = event_observed[b]

            if valid_comparison(time_a, time_b, event_a, event_b):
                paircount += 1.0
                csum += concordance_value(time_a ,time_b, pred_a, pred_b)

    if paircount == 0:
        raise ZeroDivisionError("No admissable pairs in the dataset.")
    return csum / paircount



from ifree import processing
from ifree import dicom
import os
import pickle
import pandas as pd
import numpy as np
import itk
def readNII(dcm_path):
    ImageType = itk.Image[itk.F, 3]
    reader = itk.ImageSeriesReader[ImageType].New()
    reader.SetFileName(dcm_path)
    reader.Update()
    image3d = reader.GetOutput()
    return itk.GetArrayFromImage(image3d)

if __name__ == "__main__":

    # get MRs or CTs related to RT and copy them to new dir
    old_Dir = "../originalDicom/"
    newDir = "../new/"
    patientNames, patientIDs, id2mrs, id2rt, id2dose = dicom.archiveFiles(old_Dir)
    pickle.dump((patientNames, patientIDs, id2mrs, id2rt, id2dose), open(newDir+"archive.pkl", "wb"))
    #data = pickle.load(open(newDir+"archive.pkl", "rb"))
    #patientNames, patientIDs, id2mrs, id2rt, id2dose = data


    # crop ROI-MASK and its MR
    newDir = "../new/MRandMask/"
    newSize = None # leave None to get origin size
    idx_done_list = dicom.cropROI(id2mrs, id2rt, "GTV", newDir, newSize)
    pickle.dump(idx_done_list, open(newDir+"MRandMask.pkl", "wb"))
    #idx_done_list = pickle.load(open(newDir+"MRandMask.pkl", "rb"))



    # get MR-ROI radiomics feature
    newDir = "../new/MRandMask/"
    idx = os.listdir(newDir)
    imagePaths = [newDir + i +"/mr.nii" for i in idx]
    maskPaths = [newDir + i +"/mask.nii" for i in idx]
    paramPath = "./radiomics.yaml"
    outputPath = "../new/MRandMask_features.csv"
    helper = processing.FeatureExtractor(idx, imagePaths, maskPaths, paramPath, outputPath)
    helper.extract(force=True)


    # crop ROI-MASK and its Dose
    newDir = "../new/DOSEandMask/"
    newSize = None # leave None to get origin size
    idx_done_list = dicom.cropDose(id2dose, id2rt, "GTV", newDir, newSize)
    pickle.dump(idx_done_list, open(newDir+"DOSEandMask.pkl", "wb"))
    #idx_done_list = pickle.load(open(newDir+"DOSEandMask.pkl", "rb"))

    # Get Dose-Volume Histogram (DVH)
    newDir = "../new/"
    s = "BODY"
    dose_features = {}
    selected_position_list = [s, s.lower(), s.upper(), s.capitalize(), s.title(), s.replace("-", " ")]
    for i, (pid, rtssfile) in enumerate(id2rt.items()):
        print(f"{i}/{len(id2rt)}")
        rtdosefile = id2dose[pid]
        # percents is the space between 0-100
        data, percents = dicom.get_dvh_of_key(rtssfile, rtdosefile, s)
        for roiname in selected_position_list:
            if roiname in data.keys():
                selected_position = roiname
        dose_list, max_value, min_value, mean_value = data[selected_position]
        tmp = {"pid": pid, "dose_max": max_value, "dose_min": min_value, "dose_mean": mean_value,}
        for j, pc in enumerate(percents):
            tmp[f"dose_{pc}"] = dose_list[j]
        dose_features[i] = tmp
    df_dose_features = pd.DataFrame.from_dict(dose_features, orient='index')
    df_dose_features.to_csv(newDir+"dvh_BODY.csv", index=False)
    # df_dose_features.head()


import os, csv, six
import numpy as np
import pandas as pd
from glob import glob
from tqdm import tqdm
from minepy import MINE
from scipy import stats
from itertools import permutations, combinations
from sklearn import feature_selection
from sklearn import ensemble
from sklearn import linear_model

from sklearn.decomposition import PCA
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import StandardScaler, Normalizer, MinMaxScaler

from tqdm import tqdm
from radiomics import featureextractor
import SimpleITK as sitk

class FeatureExtractor():
    def __init__(self, idList, imagePaths, maskPaths, paramPath, outputPath):
        self.idList = idList
        self.imagePaths = imagePaths
        self.maskPaths = maskPaths
        self.paramPath = paramPath
        self.outputPath = outputPath
        assert len(self.imagePaths) == len(self.maskPaths), "#imagePaths #maskPaths is not consistent!"

    def save2csv(self, features, idx):
        assert len(features) == len(idx), "#features #idx is not consistent!"

        if len(features) == 0:
            print("No features!")
        else:
            obs_header = ["id"]
            obs_header.extend(list(features[0].keys()))
            if os.path.exists(self.outputPath) or os.path.isfile(self.outputPath):
                os.remove(self.outputPath)
            f = open(self.outputPath, 'w')
            f.write(','.join(obs_header) + '\n')
            f.close()
            w = csv.DictWriter(open(self.outputPath, 'a'), fieldnames=obs_header, quoting=csv.QUOTE_MINIMAL)
            for i in range(len(features)):
                fea = features[i]
                fea["id"] = idx[i]
                w.writerows([fea])
            print("features saved to ", self.outputPath)

    def singleExtract(self, imageName, maskName, params):
        if imageName.split(".")[-1] in ["bmp", "jpg", "png"]:
            return self.singleExtract_BMP(imageName, maskName, params)
        else:
            return self.singleExtract_NII(imageName, maskName, params)

    def singleExtract_BMP(self, imageName, maskName, params):

        try:
            extractor = featureextractor.RadiomicsFeatureExtractor(params)


            color_channel = 0
            im = sitk.ReadImage(imageName)
            selector = sitk.VectorIndexSelectionCastImageFilter()
            selector.SetIndex(color_channel)
            im = selector.Execute(im)


            color_channel = 0
            mas = sitk.ReadImage(maskName)
            selector = sitk.VectorIndexSelectionCastImageFilter()
            selector.SetIndex(color_channel)
            mas = selector.Execute(mas)

            result = extractor.execute(im, mas)
            feature = {key: val for key, val in six.iteritems(result)}

            
        except Exception as e:
            print(e)
            print("error when extacting ", imageName)
            feature = None
        return feature
    def singleExtract_NII(self, imageName, maskName, params):

        try:
            extractor = featureextractor.RadiomicsFeatureExtractor(params)
            result = extractor.execute(imageName, maskName)
            feature = {key: val for key, val in six.iteritems(result)}

        except Exception as e:
            print(e)
            print("error when extacting ", imageName)
            feature = None
        return feature

    def extract(self, force=False):
        if os.path.exists(self.outputPath) and not force:
            print("file already exist : ", self.outputPath)
        else:
            features = []
            idx = []
            lens = len(self.imagePaths)
            for i in tqdm(range(lens), total=lens):
                imageName = self.imagePaths[i]
                maskName = self.maskPaths[i]
                feature = self.singleExtract(imageName, maskName, self.paramPath)
                if feature is not None:
                    features.append(feature)
                    idx.append(self.idList[i])
            self.save2csv(features, idx)




class FeatureProcess():
    def __init__(self, X_train, X_test):
        self.X_train = X_train
        self.X_test = X_test

    def simpleImpute(self, strategy='mean'):
        imp = SimpleImputer(missing_values=np.nan, strategy=strategy)
        imp.fit(self.X_train)
        self.X_train = imp.transform(self.X_train)
        self.X_test = imp.transform(self.X_test)
        
    def standardScale(self):
        scaler = StandardScaler()
        scaler.fit(self.X_train)
        self.X_train = scaler.transform(self.X_train)
        self.X_test = scaler.transform(self.X_test)

    def normalizer(self):
        nor = Normalizer()
        nor.fit(self.X_train)
        self.X_train = nor.transform(self.X_train)
        self.X_test = nor.transform(self.X_test)

    def minMaxScaler(self):
        mms = MinMaxScaler()
        mms.fit(self.X_train)
        self.X_train = mms.transform(self.X_train)
        self.X_test = mms.transform(self.X_test)
    
    def pca(self, n_components=10):
        pca = PCA(n_components=n_components)
        pca.fit(self.X_train)
        self.X_train = pca.transform(self.X_train)
        self.X_test = pca.transform(self.X_test)




class bicluster:
    def __init__(self, vec, left=None,right=None,distance=0.0,id=None) :
        self.left = left
        self.right = right
        self.vec = vec
        self.id = id
        self.distance = distance

class FeatureSelector():
    def __init__(self, features, labels, featureNames, eps=1e-3):
        self.x = features
        self.y = labels
        self.featureNames = np.array(featureNames)
        self.eps = eps
        self.n = self.x.shape[1]
        self.indexs = None

    ############################
    #   univariate selection   #
    ############################
    def univarSelector(self, top_k=1, method_name="f_classif", inplace=True):
        """
        :method_name {"chi2", "f_classif", "mutual_info_classif"}
        """
        print("Feature selecting method: ", method_name)
        selector = {"chi2": feature_selection.chi2,
                    "f_classif": feature_selection.f_classif,
                    "mutual_info_classif": feature_selection.mutual_info_classif}
        func = selector[method_name]
        sler = feature_selection.SelectKBest(func, k=top_k)
        sler.fit(self.x, self.y)
        self.indexs = sler.get_support()

        if inplace:
            self.x = self.x[:, self.indexs]
            self.n = self.x.shape[1]
            self.featureNames = self.featureNames[self.indexs]
            return self.x, self.y
        else:
            return self.x[:, self.indexs], self.y

    #########################
    # regression selecting  #
    #########################
    def lrSelector(self, method_name="lr", inplace=True):
        """
        :method_name {"lr", "ridge"}
        """
        print("Feature selecting method: ", method_name)
        selector = {"lr": linear_model.LinearRegression(), "ridge": linear_model.Ridge()}
        lr = selector[method_name]
        lr.fit(self.x, self.y)
        coefs = lr.coef_.tolist()
        self.indexs = [i for i in range(len(coefs)) if np.abs(coefs[i]) > self.eps]
        if inplace:
            self.x = self.x[:, self.indexs]
            self.n = self.x.shape[1]
            self.featureNames = self.featureNames[self.indexs]
            return self.x, self.y
        else:
            return self.x[:, self.indexs], self.y

    ############################
    #      model selecting     #
    ############################
    def modelSelector(self, model_name="rf", inplace=True):
        """
        :method_name {"rf", "lasso"}
        """
        print("Feature selecting method: ", model_name)
        selector = {"rf": ensemble.RandomForestClassifier(n_estimators=10), "lasso": linear_model.LassoCV(cv=5, max_iter=5000)}
        model = selector[model_name]
        sler = feature_selection.SelectFromModel(model)
        sler.fit(self.x, self.y)
        self.indexs = sler.get_support()

        if inplace:
            self.x = self.x[:, self.indexs]
            self.n = self.x.shape[1]
            self.featureNames = self.featureNames[self.indexs]
            return self.x, self.y
        else:
            return self.x[:, self.indexs], self.y


    #########################
    # correlation selecting #
    #########################
    def calMic(self, x1, x2):
        # Maximal Information Coefficient
        mine = MINE(alpha=0.6, c=15)
        mine.compute_score(x1, x2)
        return mine.mic()

    def hcluster(self, X, calDistance) :
        biclusters = [ bicluster(vec = X[:, i], id = i ) for i in range(X.shape[1]) ]
        distances = {}
        flag = None
        currentclusted = -1
        print("features dim: ", len(biclusters))
        while(len(biclusters) > 1) :
            max_val = -1
            biclusters_len = len(biclusters)
            for i in range(biclusters_len-1) :
                for j in range(i + 1, biclusters_len) :
                    if distances.get((biclusters[i].id,biclusters[j].id)) == None:
                        distances[(biclusters[i].id,biclusters[j].id)], _ = calDistance(biclusters[i].vec,biclusters[j].vec)
                    d = distances[(biclusters[i].id,biclusters[j].id)] 
                    if d > max_val :
                        max_val = d
                        flag = (i,j)
            bic1,bic2 = flag
            newvec = (biclusters[bic1].vec + biclusters[bic2].vec) / 2
            newbic = bicluster(newvec, left=biclusters[bic1], right=biclusters[bic2], distance=max_val, id = currentclusted)
            currentclusted -= 1
            del biclusters[bic2]
            del biclusters[bic1]
            biclusters.append(newbic)
        return biclusters[0]

    def corrSelector(self, method_name="pearson", num=1, inplace=True):
        """
        :method_name {"pearson", "kendall", "spearman", "mic"}
        """
        print("Feature selecting method: ", method_name)
        selector = {"pearson": stats.pearsonr, "kendall": stats.kendalltau,
                    "spearman": stats.spearmanr, "mic": self.calMic}
        func = selector[method_name]
        root_node = self.hcluster(self.x, func)
        node_ids = []
        nodes = [root_node]
        while(len(nodes)>0):
            tmp = nodes.pop(0)
            if tmp.id<0:
                nodes.append(tmp.left)
                nodes.append(tmp.right)
            else:
                node_ids.append(tmp.id)
        self.indexs = np.asarray(node_ids)[:num]

        if inplace:
            self.x = self.x[:, self.indexs]
            self.n = self.x.shape[1]
            self.featureNames = self.featureNames[self.indexs]
            return self.x, self.y
        else:
            return self.x[:, self.indexs], self.y

    #########################
    #    T test selecting   #
    #########################
    def calTtest(self, x1, x2, threshold=0.05):
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.ttest_ind.html
        stat, p = stats.levene(x1, x2)
        if p > threshold:
            _, t = stats.ttest_ind(x1, x2)
        else:
            _, t = stats.ttest_ind(x1, x2, equal_var=False)
        return t

    def ttestSelector(self, threshold=0.05, inplace=True):
        print("Feature selecting method: T test")
        labels = np.unique(self.y)
        num_labels = len(labels)
        cbs = list(combinations(range(num_labels), 2))
        num_cbs = len(cbs)
        eps = 0.1
        self.indexs = []
        for i in range(self.n):
            count = 0
            for c in range(num_cbs):
                index1, index2 = cbs[c]
                row1, row2 = labels[index1], labels[index2]
                r1, r2 = self.y==row1, self.y==row2
                t = self.calTtest(self.x[r1, i], self.x[r2, i])
                if t > threshold:
                    count += 1
            if count/num_cbs < eps:
                self.indexs.append(i)

        if inplace:
            self.x = self.x[:, self.indexs]
            self.n = self.x.shape[1]
            self.featureNames = self.featureNames[self.indexs]
            return self.x, self.y
        else:
            return self.x[:, self.indexs], self.y

    #########################
    #   mann whitney u test #
    #########################
    def mannSelector(self, threshold=0.05, inplace=True):
        print("Feature selecting method: mann whitney u test")
        labels = np.unique(self.y)
        num_labels = len(labels)
        cbs = list(combinations(range(num_labels), 2))
        num_cbs = len(cbs)
        eps = 0.1
        self.indexs = []
        for i in range(self.n):
            count = 0
            for c in range(num_cbs):
                index1, index2 = cbs[c]
                row1, row2 = labels[index1], labels[index2]
                r1, r2 = self.y==row1, self.y==row2
                _, t = stats.mannwhitneyu(self.x[r1, i], self.x[r2, i], alternative='two-sided')
                if t > threshold:
                    count += 1
            if count/num_cbs < eps:
                self.indexs.append(i)

        if inplace:
            self.x = self.x[:, self.indexs]
            self.n = self.x.shape[1]
            self.featureNames = self.featureNames[self.indexs]
            return self.x, self.y
        else:
            return self.x[:, self.indexs], self.y

    #########################
    #       combination     #
    #########################
    def combination(self, train_test_run, x_test, y_test, inplace=True):
        print("Feature selecting method: combination")
        len_f = self.x.shape[1]
        res_indexs = []
        min_value = 2
        for c_f in range(1,len_f):
            cbs = list(combinations(list(range(len_f)), c_f))
            for c in cbs:
                indexs = [i for i in c]
                error_mse, error_mae = train_test_run(self.x[:, indexs], self.y, x_test[:, indexs], y_test)
                if error_mae<min_value:
                    min_value=error_mae
                    self.indexs = indexs
                    res_indexs.append({"error_mse":error_mse,"error_mae":error_mae,"indexs":indexs})
        print(res_indexs[-1])
        if inplace:
            self.x = self.x[:, self.indexs]
            self.n = self.x.shape[1]
            self.featureNames = self.featureNames[self.indexs]
            return self.x, self.y
        else:
            return self.x[:, self.indexs], self.y



import os
import cv2
import itk
import vtk
import shutil
import pickle
import pydicom
import argparse
import numpy
import numpy as np
import pandas as pd
from tqdm import tqdm
import SimpleITK as sitk
from dicompylercore import dvhcalc
from dicompylercore import dicomparser
from vtk.util.numpy_support import vtk_to_numpy

ImageType   =  itk.Image[itk.F, 3]
ImageTypeUC =  itk.Image[itk.UC,3]

def reorientation(input_image):
    region3d = input_image.GetBufferedRegion()
    start3d = region3d.GetIndex()

    # region3d = mask3d.GetBufferedRegion()
    # start3d = region3d.GetIndex()
    # size3d = region3d.GetSize()
    spacing3d = input_image.GetSpacing()
    #new_direction_list = [1., 0., 0., 0., 1., 0., 0., 0., 1.]
    new_direction_list = [[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]]
    new_direction = itk.matrix_from_array(np.array(new_direction_list))
    #print(new_direction)
    

    # def is_equal(old, new, EPS=1e-6):
    #     for a, b in zip(old, new):
    #         if np.abs(a-b) > EPS:
    #             return False
    #     return True
    # if np.array_equal(old_direction, new_direction):
    #     meta = {}
    #     meta['origin'] = input_image.GetOrigin()
    #     meta['direction'] = old_direction
    #     return input_image, meta

    I, J, K = input_image.GetBufferedRegion().GetSize()
    coords = []
    for i in [0, I-1]:
        for j in [0, J-1]:
            for k in [0, K-1]:
                coords.append([i, j, k])
    coords = [input_image.TransformIndexToPhysicalPoint(t) for t in coords]
    coords = np.array(coords)
    min_i, max_i = np.min(coords[:, 0]), np.max(coords[:, 0])
    min_j, max_j = np.min(coords[:, 1]), np.max(coords[:, 1])
    min_k, max_k = np.min(coords[:, 2]), np.max(coords[:, 2])
    new_origin = [min_i, min_j, min_k]

    new_size = [0, 0, 0]
    new_size[0] = int(round((max_i-min_i)/spacing3d[0]))+1
    new_size[1] = int(round((max_j-min_j)/spacing3d[1]))+1
    new_size[2] = int(round((max_k-min_k)/spacing3d[2]))+1

    # resampleFilter = sitk.ResampleImageFilter()
    # resampleFilter.SetOutputSpacing(spacing3d)
    # resampleFilter.SetOutputOrigin(new_origin)
    # resampleFilter.SetOutputDirection(new_direction) ####
    # resampleFilter.SetInterpolator(sitk.sitkNearestNeighbor)
    # resampleFilter.SetSize(new_size)
    # output_image = resampleFilter.Execute(input_image)

    # resample
    interpolatorType = itk.LinearInterpolateImageFunction[ImageType, itk.D]
    interpolator = interpolatorType.New()
    resample2dFilter = itk.ResampleImageFilter[ImageType, ImageType].New()
    resample2dFilter.SetInterpolator(interpolator)
    resample2dFilter.SetSize(new_size)
    resample2dFilter.SetOutputOrigin(new_origin)
    resample2dFilter.SetOutputDirection(new_direction)
    resample2dFilter.SetOutputSpacing(spacing3d)
    resample2dFilter.SetOutputStartIndex(start3d)
    resample2dFilter.SetInput(input_image)
    resample2dFilter.Update()
    output_image = resample2dFilter.GetOutput()

    meta = {}
    meta['origin'] = new_origin
    meta['direction'] = new_direction
    return output_image
    #return output_image, meta

#########################
#####    读写NII     ####
#########################
def writeNii(write_path, Image, imageType):
    # write_path ".nii"
    writer = itk.ImageFileWriter[imageType].New()
    writer.SetInput(Image)
    writer.SetFileName(write_path)
    writer.Update()


#########################
#####   读写DICOM    ####
#########################
def readDicom(dcm_path, imageType):
    # 读取dicom文件
    reader = itk.ImageSeriesReader[imageType].New()
    dicomIO = itk.GDCMImageIO.New()
    reader.SetImageIO(dicomIO)
    reader.SetFileNames(dcm_path)
    reader.Update()
    image3d = reader.GetOutput()

    # 特定的窗宽窗位
    windowWidth, windowLevel, minValue = ImageAutoWidthLevel3DFuction(image3d)
    windowLevelFilter = itk.IntensityWindowingImageFilter[imageType, imageType].New()
    windowLevelFilter.SetInput(image3d)
    windowLevelFilter.SetOutputMaximum(350)
    windowLevelFilter.SetOutputMinimum(0)
    windowLevelFilter.SetWindowLevel(float(windowWidth), float(windowLevel))
    windowLevelFilter.Update()

    return windowLevelFilter.GetOutput()

def readDoseDicom(dcm_path, imageType):
    # 读取dicom文件
    reader = itk.ImageSeriesReader[imageType].New()
    dicomIO = itk.GDCMImageIO.New()
    reader.SetImageIO(dicomIO)
    reader.SetFileName(dcm_path)
    reader.Update()
    image3d = reader.GetOutput()

    return image3d

def ReadRTDicom(rtPath,image,ContourList):
    pointXYZ = []
    UNReadALL = False
    if len(ContourList) != 0 :
        UNReadALL = True
        ContourList = [string.upper() for string in ContourList]

    # origin spacing
    origin  = image.GetOrigin()
    spacing = image.GetSpacing()
    size    = image.GetLargestPossibleRegion().GetSize()
    index   = image.GetLargestPossibleRegion().GetIndex()
    # get the rotation matrix and the inverse
    directionMatrix = itk.GetArrayFromVnlMatrix(image.GetDirection().GetVnlMatrix().as_matrix())
    rotMat    = np.zeros(9)
    rotMatInv = np.zeros(9)
    for i in range(3):
        for j in range(3):
            rotMat[i*3 + j] = directionMatrix[i, j]
            rotMatInv[i*3 +j] = directionMatrix[j, i]

    #read RT
    rtdcm_ds = pydicom.read_file(rtPath, force=True)
    #校验 RT
    try:
        ROIContourSequenceA      = rtdcm_ds[0x3006, 0x0039].value
        StructureSetROISequenceA = rtdcm_ds[0x3006, 0x0020].value
    except Exception:
        print('Problem locating [0x3006,0x0020],[0x3006, 0x0039] - Is this a valid RT Structure file? ')
        return

    # loop through structures
    ROI = {}
    DirOfNumStructureSetROIA = {}
    DirOfNumROIContourA = {}
    for StructureSetROI,ROIContour in zip(StructureSetROISequenceA,ROIContourSequenceA):
        try:
            DirOfNumStructureSetROIA[StructureSetROI[0x3006, 0x0022].value] = StructureSetROI
            DirOfNumROIContourA     [ROIContour     [0x3006, 0x0084].value] = ROIContour
        except Exception:
            print('Problem locating [0x3006, 0x0022],[0x3006, 0x0084] - Is this a valid RT Structure file? ')
            return

    for ROIContour in ROIContourSequenceA:
        try:
            StructureSetROI = DirOfNumStructureSetROIA[ROIContour[0x3006, 0x0084].value]
            OrganName = StructureSetROI[0x3006, 0x0026].value
            if UNReadALL and (OrganName.upper() not in ContourList):
                continue
        except Exception:
            print('Problem locating [0x3006, 0x0026],[0x3006, 0x0084] - Is this a valid RT Structure file? ')
            return

        # vtkPoly contours
        contours = vtk.vtkPolyData()
        # now loop through each item or this structure
        # eg one prostate region on a single slice is an item
        cell = vtk.vtkCellArray()
        points = vtk.vtkPoints()
        pointId = 0

        try:
            # loop for this organ
            for roi in ROIContour[0x3006, 0x0040].value:
                polyLine = vtk.vtkPolyLine()
                data = np.array(roi[0x3006, 0x0050].value)

                npts = len(data) // 3

                for j in range(npts):
                    p = [float(data[j * 3 + 0]), float(data[j * 3 + 1]), float(data[j * 3 + 2])]
                    # consider the direction
                    p = linearTransform(rotMat, origin, p)
                    pointXYZ.append(p)
                    pointId = points.InsertNextPoint(p)
                    polyLine.GetPointIds().InsertNextId(pointId)

                # start connected to end
                p = [float(data[0]), float(data[1]), float(data[2])]
                # consider the direction
                p = linearTransform(rotMat, origin, p)
                pointId = points.InsertNextPoint(p)
                polyLine.GetPointIds().InsertNextId(pointId)
                cell.InsertNextCell(polyLine)
        except Exception as e:
            print(e)
            print('Problem locating [0x3006, 0x0050],[0x3006, 0x0040] - Is this a valid RT Structure file? ')
            return

        contours.SetPoints(points)
        contours.SetLines(cell)
        contours.Modified()

        # get bounds
        bounds = list(contours.GetBounds())
        # roi origin spacing size for X
        Xmin = numpy.floor((bounds[0] - origin[0]) / spacing[0]) - 1
        Xmax = numpy.floor((bounds[1] - origin[0]) / spacing[0]) + 2
        Xextent = Xmax - Xmin
        originX = Xmin * spacing[0] + origin[0]

        # roi origin spacing size for Y
        Ymin = numpy.floor((bounds[2] - origin[1]) / spacing[1]) - 1
        Ymax = numpy.floor((bounds[3] - origin[1]) / spacing[1]) + 2
        Yextent = Ymax - Ymin
        originY = Ymin * spacing[1] + origin[1]

        # roi origin spacing size for Z
        Zmin = numpy.floor((bounds[4] - origin[2]) / spacing[2]) - 1
        Zmax = numpy.floor((bounds[5] - origin[2]) / spacing[2]) + 2
        Zextent = Zmax - Zmin
        originZ = Zmin * spacing[2] + origin[2]

        originRoi = [originX, originY, originZ]
        sizeRoi = [Xextent + 1, Yextent + 1, Zextent + 1]

        # convert the contours to itkImage
        #print('contours points number: ', contours.GetNumberOfPoints())
        #print('contours cells number: ', contours.GetNumberOfCells())
        if contours.GetNumberOfPoints() == 0:
            continue

        # prepare the binary image 's voxel grid
        whiteImage = vtk.vtkImageData()
        whiteImage.SetSpacing(spacing)
        whiteImage.SetExtent(0, int(sizeRoi[0]) - 1, 0, int(sizeRoi[1]) - 1, 0, int(sizeRoi[2]) - 1)
        whiteImage.SetOrigin(originRoi)
        whiteImage.AllocateScalars(vtk.VTK_UNSIGNED_CHAR, 1)

        # vtkIdType
        count = whiteImage.GetNumberOfPoints()
        whiteImage.GetPointData().GetScalars().Fill(1)

        # polygonal data --> image stencil:
        pol2stenc = vtk.vtkPolyDataToImageStencil()
        pol2stenc.SetTolerance(0)  # important if extruder->SetVector(0, 0, 1) !!!
        pol2stenc.SetInputData(contours)
        pol2stenc.SetOutputOrigin(originRoi)
        pol2stenc.SetOutputSpacing(spacing)
        pol2stenc.SetOutputWholeExtent(whiteImage.GetExtent())
        pol2stenc.Update()

        # cut the corresponding white image and set the background:
        imgstenc = vtk.vtkImageStencil()
        imgstenc.SetInputData(whiteImage)
        imgstenc.SetStencilData(pol2stenc.GetOutput())
        imgstenc.ReverseStencilOff()
        imgstenc.SetBackgroundValue(0)
        imgstenc.Update()

        imageOrganVtk = imgstenc.GetOutput()

        # vtk image to numpy and to itk
        rows, cols, pages = imageOrganVtk.GetDimensions()
        imageNp = vtk_to_numpy(imageOrganVtk.GetPointData().GetScalars())
        imageNp = numpy.ascontiguousarray(imageNp.reshape(pages, cols, rows) * 255)
        # assert imageNp.shape == imageOrganVtk.GetDimensions()
        imageOrganItk = itk.GetImageFromArray(imageNp)
        imageOrganItk.SetSpacing(imageOrganVtk.GetSpacing())
        # for origin direction is considered
        originRoi = imageOrganVtk.GetOrigin()
        originRoi = linearTransform(rotMatInv, origin, originRoi)
        imageOrganItk.SetOrigin(originRoi)
        # direction
        imageOrganItk.SetDirection(image.GetDirection())
        #  image push back
        ROI[OrganName] = imageOrganItk

    return ROI, pointXYZ


#########################
#####  获取文件地址   ####
#########################
def GetFilePath(ImagePath):
    ctfiles = []
    mrfiles = []
    z = []
    z_mr = []
    patientID = ''
    patientName = ''
    rtfile = ''
    dosefile = ''
    for file_path in os.listdir(ImagePath):
        if not file_path.lower().endswith('dcm'):
            print('Warn, {} is not a dicom file'.format(file_path))
            continue
        dcm = pydicom.read_file(ImagePath + '/' + file_path, force=True)
        try:
            dcm.SOPClassUID
            patientName = dcm.PatientName
        except:
            print('file {} is corrupted, pass'.format(file_path))
            continue
        # RS rt
        if dcm.SOPClassUID == '1.2.840.10008.5.1.4.1.1.481.3':
            rtfile = ImagePath + '/' + file_path
            patientID = dcm.PatientID
        # ct
        elif dcm.SOPClassUID == '1.2.840.10008.5.1.4.1.1.2':
            ctfiles.append(ImagePath + '/' + file_path)
            z.append(dcm.SliceLocation)
        # mr
        elif dcm.SOPClassUID == '1.2.840.10008.5.1.4.1.1.4':
            mrfiles.append(ImagePath + '/' + file_path)
            z_mr.append(dcm.SliceLocation)
        # RD dose
        elif dcm.SOPClassUID == '1.2.840.10008.5.1.4.1.1.481.2':
            dosefile = ImagePath + '/' + file_path
        # RP
        elif dcm.SOPClassUID == '1.2.840.10008.5.1.4.1.1.481.5':
            pass
    if len(ctfiles)>0:
        zipped = zip(z, ctfiles)
        sort_zipped = sorted(zipped, key=lambda x: x[0])
        result = zip(*sort_zipped)
        z, ctfiles = [list(x) for x in result]

    if len(mrfiles)>0:
        zipped2 = zip(z_mr, mrfiles)
        sort_zipped2 = sorted(zipped2, key=lambda x: x[0])
        result2 = zip(*sort_zipped2)
        z_mr, mrfiles = [list(x) for x in result2]

    return ctfiles, rtfile, mrfiles, dosefile, patientID, patientName


#########################
#####  获取窗宽窗位   ####
#########################
def ImageAutoWidthLevel3DFuction(image3d):
    minValue = 0
    maxValue = 1
    windowLevel = 0
    windowWidth = 0

    npImage = itk.GetArrayFromImage(image3d)

    minValue = npImage.min()
    maxValue = npImage.max()

    # if gray level less than 500
    if (maxValue - minValue < 500):
        windowLevel = np.floor((minValue + maxValue) / 2.0)
        windowWidth = maxValue - minValue
        return windowWidth, windowLevel, minValue

    # parameters for histogram
    binNum = 500
    lowPercent = 0.04
    upPercent = 0.999
    lowWindow = minValue
    upWindow = maxValue
    windowLevel = (minValue + maxValue) / 2.0
    windowWidth = maxValue - minValue

    # image information
    imageSize = image3d.GetLargestPossibleRegion().GetSize()

    # ImageToHistogramFilter
    npImage = itk.GetArrayFromImage(image3d)
    histo, edge = np.histogram(npImage, bins=binNum, range=(lowWindow, upWindow))

    # summation to compare
    tempAddSum = 0
    tpHistogramNum = histo.size
    tempTotalNum = imageSize[0] * imageSize[1] * imageSize[2]

    # find the lower value
    for i in range(tpHistogramNum):

        tempAddSum += histo[i]
        if (tempAddSum > tempTotalNum * lowPercent):
            lowWindow = i * (maxValue - minValue) / (1.0 * binNum) + minValue;
            break

    # find the upper value
    tempAddSum = 0
    for i in range(tpHistogramNum):

        tempAddSum += histo[tpHistogramNum - i - 1]
        if (tempAddSum > tempTotalNum * (1 - upPercent)):
            upWindow = maxValue - i * (maxValue - minValue) / (1.0 * binNum);
            break

    # final we get the W/L
    windowWidth = upWindow - lowWindow
    windowLevel = lowWindow + windowWidth / 2
    return windowWidth, windowLevel, minValue


def linearTransform(rotMat, origin, coord):
    """
    Linear transform $3\times1$ vector to a new coordinate system given by
    $3\times3$ rotMat (in row) and a $3\times 1$ origin. If see rotMat's
    rows as base vectors of a new coord. system, this transform is mapping old
    system to new system!
    :param rotMat: $ 3x3 rotation matrix in row expansion
    :param origin: $ 3x1 rotation origin
    :param coord: 3x1 array in old coordinate system
    :return: 3x1 array in new coordinate system
    """
    nodeX = rotMat[0] * (coord[0] - origin[0]) + rotMat[1] * (coord[1] - origin[1]) + \
            rotMat[2] * (coord[2] - origin[2]) + origin[0]
    nodeY = rotMat[3] * (coord[0] - origin[0]) + rotMat[4] * (coord[1] - origin[1]) + \
            rotMat[5] * (coord[2] - origin[2]) + origin[1]
    nodeZ = rotMat[6] * (coord[0] - origin[0]) + rotMat[7] * (coord[1] - origin[1]) + \
            rotMat[8] * (coord[2] - origin[2]) + origin[2]
    return [nodeX, nodeY, nodeZ]


#########################
### 识别单个患者图像信息 ##
#########################
def GetRTPath(ImagePath):
    patientID = ''
    rtfile = ''
    for file_path in os.listdir(ImagePath):
        if not file_path.lower().endswith('dcm'):
            print('Warn, {} is not a dicom file'.format(file_path))
            continue
        dcm = pydicom.read_file(ImagePath + '/' + file_path, force=True)
        try:
            dcm.SOPClassUID
        except:
            print('file {} is corrupted, pass'.format(file_path))
            continue
        # RS rt
        if dcm.SOPClassUID == '1.2.840.10008.5.1.4.1.1.481.3':
            rtfile = ImagePath + '/' + file_path
            patientID = dcm.PatientID
            patientName = dcm.PatientName
            return rtfile, patientID, patientName
    return None

def list_dcm_files(directory):
    '''
        find all dicom files in a directory

        :param directory: a directory contain dicom files, this directory
            can contain sub-directories
        :return: a list, which each item contrain dcm's SOPInstanceUID, and value contain
            path, SOPClassUID
    '''
    dcm_files = []
    for root, dirs, fns in os.walk(directory):
        for fn in fns:
            dcm_fn = os.path.join(root, fn)
            dcm = pydicom.read_file(dcm_fn, force=True)
            if hasattr(dcm, 'SOPClassUID'):
                if not hasattr(dcm, 'SOPInstanceUID'):
                    print(
                        '{} is a dicom file, but dose not have SOPInstanceUID'.format(dcm_fn))
                    continue
                dcm_files.append({
                    'path': dcm_fn,
                    'SOPClassUID': dcm.SOPClassUID,
                    'SOPInstanceUID': dcm.SOPInstanceUID,
                    'file': dcm,
                })
            else:
                print('{} is not a dicom file'.format(dcm_fn))
    return dcm_files


def get_structure_related_image_files(structure_dcm, dcm_files):
    '''
        find all relevant dicom images of a structure,
        1. if structure file have ContourImageSequence, use ReferencedSOPInstanceUID in structure
        2. else, if structure file have FrameOfReferenceUID, use FrameOfReferenceUID
    '''
    related_image_files = []
    if hasattr(structure_dcm, 'ReferencedFrameOfReferenceSequence'):
        ReferencedFrameOfReferenceSequence = structure_dcm.ReferencedFrameOfReferenceSequence
        if len(ReferencedFrameOfReferenceSequence) > 1:
            print('More than one ReferencedFrameOfReference found')
        if len(ReferencedFrameOfReferenceSequence) == 0:
            print(
                'Cannot find ReferencedFrameOfReference in structure')
        ReferencedFrameOfReference = ReferencedFrameOfReferenceSequence[0]
        if hasattr(ReferencedFrameOfReference, 'RTReferencedStudySequence'):
            try:
                RTReferencedStudy = ReferencedFrameOfReference.RTReferencedStudySequence[0]
                RTReferencedSeriesSequence = RTReferencedStudy.RTReferencedSeriesSequence
                RTRefenecedSeries = RTReferencedSeriesSequence[0]
                for ContourImage in RTRefenecedSeries.ContourImageSequence:
                    ReferencedSOPInstanceUID = ContourImage.ReferencedSOPInstanceUID
                    find_image = False
                    for item in dcm_files:
                        if item['SOPInstanceUID'] == ReferencedSOPInstanceUID:
                            find_image = True
                            related_image_files.append(item)
                            break
                    if not find_image:
                        print('Cannot find {} which referenced by structure'.format(
                            ReferencedSOPInstanceUID))
            except:
                pass
        else:
            if hasattr(ReferencedFrameOfReference, 'FrameOfReferenceUID'):
                FrameOfReferenceUID = ReferencedFrameOfReference.FrameOfReferenceUID
                try:
                    for item in dcm_files:
                        if item['file'].FrameOfReferenceUID == FrameOfReferenceUID:
                            related_image_files.append(item)
                except:
                    pass
    else:
        if hasattr(structure_dcm, 'FrameOfReferenceUID'):
            FrameOfReferenceUID = ReferencedFrameOfReference.FrameOfReferenceUID
            try:
                for item in dcm_files:
                    if item['file'].FrameOfReferenceUID == FrameOfReferenceUID:
                        related_image_files.append(item)
            except:
                pass
    #return related_image_files
    return sorted(related_image_files, key=lambda x: x["file"][0x0008,0x0018].value)

#########################
###  寻找与RT相关的文件  ##
#########################
def archiveFiles(mrDir, mrDir2=None, copyed=False):
    patientNames = []
    patientIDs = []
    id2mrs = {}
    id2rt = {}
    id2dose = {}
    for patient_dir in os.listdir(mrDir):
        print("================================")
        # get RT
        patienDir = os.path.join(mrDir, patient_dir)
        ctfiles, rtfile, mrfiles, dosefile, patientID, patientName = GetFilePath(patienDir)


        print("rtfile ", rtfile)
        print("dosefile ", dosefile)
        print("patientID ", patientID)
        print("patientName ", patientName)
        if rtfile=='' or dosefile=='':
            print("No RT or DOSE found!")
            continue
        # get MRs
        #dcms_all = list_dcm_files(patienDir)
        #dcms_related = get_structure_related_image_files(pydicom.read_file(rtfile), dcms_all)
        #print("MR or CT related #", len(dcms_related))
        # make new dir
        if copyed:
            new_patient_dir = os.path.join(mrDir2, str(patientID))
            os.makedirs(new_patient_dir, exist_ok=True)
            # copy RT file
            shutil.copyfile(rtfile, os.path.join(new_patient_dir, rtfile.split("/")[-1]))
            # copy Dose file
            shutil.copyfile(dosefile, os.path.join(new_patient_dir, dosefile.split("/")[-1]))
            # copy MR files
        if len(ctfiles)>0:
        #if len(dcms_related)>0:
            patientNames.append(patientName)
            patientIDs.append(patientID)
            #id2mrs[patientID] = [ dc["path"] for dc in dcms_related]
            id2mrs[patientID] = [ dc for dc in ctfiles]
            id2rt[patientID] = rtfile
            id2dose[patientID] = dosefile
            if copyed:
                for p in dcms_related:
                    shutil.copyfile(p["path"], os.path.join(new_patient_dir, p["path"].split("/")[-1]))
    return patientNames, patientIDs, id2mrs, id2rt, id2dose


def cropROI(id2mrs, id2rt, roiName, newDir, newSize=None):
    s = roiName
    selected_position_list = [s, s.lower(), s.upper()]
    ContourList = selected_position_list
    idx_done = []
    for pid, mrfiles in id2mrs.items():
        try:
            selected_position = None
            image3d = readDicom(mrfiles, ImageType)
            ROI, pointXYZ = ReadRTDicom(id2rt[pid], image3d, ContourList)

            for i in selected_position_list:
                if i in ROI.keys():
                    selected_position = i
            if selected_position is None:
                print("No roi found!")
                continue
            else:
                print(f"{selected_position} found!")
            mask3d = ROI[selected_position]

            
            origion3d = mask3d.GetOrigin()
            spacing3d = mask3d.GetSpacing()
            region3d = mask3d.GetBufferedRegion()
            start3d = region3d.GetIndex()
            size3d = region3d.GetSize()
            if newSize is None:
                newOrigin = origion3d
            else:
                newOrigin = [0,0,0]
                newOrigin[0] = origion3d[0] + spacing3d[0]*size3d[0]/2 - spacing3d[0]*newSize[0]/2
                newOrigin[1] = origion3d[1] + spacing3d[1]*size3d[1]/2 - spacing3d[1]*newSize[1]/2
                newOrigin[2] = origion3d[2] + spacing3d[2]*size3d[2]/2 - spacing3d[2]*newSize[2]/2
            
            #image3d_reorient = reorientation(image3d)
            

            # resample Mask size
            interpolatorType = itk.LinearInterpolateImageFunction[ImageTypeUC, itk.D]
            interpolator = interpolatorType.New()
            resample2dFilter = itk.ResampleImageFilter[ImageTypeUC, ImageTypeUC].New()
            resample2dFilter.SetInterpolator(interpolator)
            resample2dFilter.SetSize(newSize)
            resample2dFilter.SetOutputOrigin(newOrigin)
            resample2dFilter.SetOutputSpacing(spacing3d)
            resample2dFilter.SetOutputStartIndex(start3d)
            resample2dFilter.SetInput(mask3d)
            resample2dFilter.Update()
            reampleMask3d = resample2dFilter.GetOutput()

            # resample MR according to Mask
            interpolatorType = itk.LinearInterpolateImageFunction[ImageType, itk.D]
            interpolator = interpolatorType.New()
            resample2dFilter = itk.ResampleImageFilter[ImageType, ImageType].New()
            resample2dFilter.SetInterpolator(interpolator)
            resample2dFilter.SetSize(newSize)
            resample2dFilter.SetOutputOrigin(newOrigin)
            resample2dFilter.SetOutputSpacing(spacing3d)
            resample2dFilter.SetOutputStartIndex(start3d)
            resample2dFilter.SetInput(image3d)
            resample2dFilter.Update()
            reampleImage3d = resample2dFilter.GetOutput()

            new_patient_dir = os.path.join(newDir, str(pid))
            os.makedirs(new_patient_dir, exist_ok=True)

            # save mask
            mask_nii_path = os.path.join(new_patient_dir, "mask.nii")
            writeNii(mask_nii_path, reampleMask3d, ImageTypeUC)
            
            image_nii_path = os.path.join(new_patient_dir, "mr.nii")
            writeNii(image_nii_path, reampleImage3d, ImageType)
            idx_done.append(pid)
        except Exception as e:
            print(e)
    return idx_done 


def cropDose(id2dose, id2rt, roiName, newDir, newSize=None):
    ImageType = itk.Image[itk.F, 3]
    ImageTypeUC =  itk.Image[itk.UC,3]

    s = roiName
    selected_position_list = [s, s.lower(), s.upper(), s.capitalize(), s.title(), s.replace("-", " ")]
    ContourList = selected_position_list
    idx_done = []
    for pid, rtfile in id2rt.items():
        try:
            selected_position = None
            image3d = readDoseDicom(id2dose[pid], ImageType)
            image3d = readDicom(mrfiles, ImageType)
            ROI, pointXYZ = ReadRTDicom(rtfile, image3d, ContourList)

            for i in selected_position_list:
                if i in ROI.keys():
                    selected_position = i
            if selected_position is None:
                print("No roi found!")
                continue
            else:
                print(f"{selected_position} found!")

            mask3d = ROI[selected_position]
            
            origion3d = mask3d.GetOrigin()
            spacing3d = mask3d.GetSpacing()
            region3d = mask3d.GetBufferedRegion()
            start3d = region3d.GetIndex()
            if newSize is None:
                size3d = region3d.GetSize()
            else:
                size3d = newSize
            
            #image3d_reorient = reorientation(image3d)

            # resample Dose according to Mask
            interpolatorType = itk.LinearInterpolateImageFunction[ImageType, itk.D]
            interpolator = interpolatorType.New()
            resample2dFilter = itk.ResampleImageFilter[ImageType, ImageType].New()
            resample2dFilter.SetInterpolator(interpolator)
            resample2dFilter.SetSize(size3d)
            resample2dFilter.SetOutputOrigin(origion3d)
            resample2dFilter.SetOutputSpacing(spacing3d)
            resample2dFilter.SetOutputStartIndex(start3d)
            resample2dFilter.SetInput(image3d)
            resample2dFilter.Update()
            reampleImage3d = resample2dFilter.GetOutput()

            new_patient_dir = os.path.join(newDir, str(pid))
            if not os.path.exists(new_patient_dir):
                os.makedirs(new_patient_dir)

            # save mask
            mask_nii_path = os.path.join(new_patient_dir, "mask.nii")
            writeNii(mask_nii_path, mask3d, ImageTypeUC)
            
            image_nii_path = os.path.join(new_patient_dir, "dose.nii")
            writeNii(image_nii_path, reampleImage3d, ImageType)
            idx_done.append(pid)
        except Exception as e:
            print(e)
    return idx_done 
#########################
###     获取dvh数据     ##
#########################
def get_dvh_of_key(rtssfile, rtdosefile, keyname=None):
    percents = [0, 1, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 97, 99]
    RTss = dicomparser.DicomParser(rtssfile)
    RTstructures = RTss.GetStructures()

    keyname_list = [keyname, keyname.lower(), keyname.upper(), keyname.capitalize(), keyname.title(), keyname.replace("-", " ")]

    # Generate the calculated DVHs
    # array_counts = []
    # array_bins = []
    data = {}
    for key, structure in RTstructures.items():
        #print(structure)
        roiname = structure["name"]
        if (keyname is not None) and (roiname not in keyname_list):
            continue
        res = dvhcalc.get_dvh(rtssfile, rtdosefile, key)
        if (len(res.counts) and res.counts[0] != 0):
            # print('DVH found for ' + keyname)
            # array_counts.append( res.counts) 
            # array_bins.append( res.bins)
            doses = [res.statistic(f"D{i}").value for i in percents]
            data[roiname] = (doses, res.max, res.min, res.mean)

    return data, percents


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages


with open('README.md', 'r') as readme_file:
    readme = readme_file.read()

requirements = [
    "dicom2nifti",
    "dicompyler-core",
    "lifelines",
    "itk",
    "minepy",
    "numpy",
    "numpy>=1.15.4",
    "pandas>=0.24.2",
    "pyradiomics",
    "scikit-image>=0.14",
    "scipy",
    "SimpleITK",
    "sklearn",
    "tqdm",
    "vtk",
]


setup(
    name='ifree',
    version='1.1.1',
    description="i love freedom, free my hand.",
    long_description=readme,
    long_description_content_type='text/markdown',
    author="Lin Zhenyu",
    author_email='linzhenyu1996@gmail.com',
    url='https://github.com/linzhenyuyuchen/ifree',
    packages=find_packages(include=['ifree']),
    include_package_data=True,
    install_requires=requirements,
    license="BSD license",
    zip_safe=False,
    keywords='ifree',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.6',
    ],
    python_requires='>=3.6'
)

